/*
 * Copyright (C) 2000-2004 Clemens Fuchslocher <clemens@allesdurcheinander.de>
 *
 * tcptunnel - 2004-12-28 - v0.2
 *             2000-09-14 - v0.1g
 *             2000-09-11 - v0.1f
 *             2000-08-15 - v0.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>

#include <time.h>

#ifdef __MINGW32__
#include <winsock2.h>
#else
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include "tcptunnel.h"

struct struct_rc rc;
struct struct_options options;
struct struct_settings settings = { 0, 0, 0, 0, 0 };

static struct option long_options[] = {
	{ "local-port",    required_argument, NULL, LOCAL_PORT_OPTION  },
	{ "remote-host",   required_argument, NULL, REMOTE_HOST_OPTION },
	{ "remote-port",   required_argument, NULL, REMOTE_PORT_OPTION },
	{ "stay-alive",    no_argument,       NULL, STAY_ALIVE_OPTION  },
	{ "help",          no_argument,       NULL, HELP_OPTION        },
	{ "version",       no_argument,       NULL, VERSION_OPTION     },
	{ "log-to-stdout", no_argument,       NULL, LOG_TO_STDOUT      },
	{ 0, 0, 0, 0 }
};

int main (int argc, char *argv[])
{
#ifdef __MINGW32__
	WSADATA info;
	if (WSAStartup (MAKEWORD(1,1), &info) != 0)
	{
		perror ("main: WSAStartup ()");
		exit (KO);
	}
#endif

	exec_name = strrchr (argv[0], PATH_SEPARATOR);
	if (exec_name)
	{
		exec_name++;
	}
	else
	{
		exec_name = argv[0];
	}

	get_commandline_param (argc, argv);

	if (build_server () == KO)
	{
		exit (KO);
	}

	do
	{
		if (wait_for_clients () == OK)
		{
			if (build_tunnel () == OK)
			{
				use_tunnel ();
			}
		}
	}
	while (settings.stay_alive);

	close (rc.server_socket);

	return OK;
}

void get_commandline_param (int argc, char *argv[])
{
	int opt;
	int opt_ind;

	do
	{
		opt = getopt_long (argc, argv, "", long_options, &opt_ind);
		switch (opt)
		{
			case LOCAL_PORT_OPTION:
			{
				set_option (&options.local_port, optarg);
				settings.local_port = 1;
				break;
			}

			case REMOTE_HOST_OPTION:
			{
				set_option (&options.remote_host, optarg);
				settings.remote_host = 1;
				break;
			}

			case REMOTE_PORT_OPTION:
			{
				set_option (&options.remote_port, optarg);
				settings.remote_port = 1;
				break;
			}

			case STAY_ALIVE_OPTION:
			{
				settings.stay_alive = 1;
				break;
			}

			case HELP_OPTION:
			{
				print_usage ();
				print_help ();
				exit (KO);
			}

			case VERSION_OPTION:
			{
				print_version ();
				exit (KO);
			}

			case LOG_TO_STDOUT:
			{
				settings.log_to_stdout = 1;
				break;
			}

			case '?':
			{
				print_usage ();
				print_helpinfo ();
				exit (KO);
			}
		}
	}
	while (opt != -1);

	if (!settings.local_port)
	{
		print_missing ("missing '--local-port=' option.");
		exit (KO);
	}
	else if (!settings.remote_host)
	{
		print_missing ("missing '--remote-host=' option.");
		exit (KO);
	}
	else if (!settings.remote_port)
	{
		print_missing ("missing '--remote-port=' option.");
		exit (KO);
	}
}

int build_server (void)
{
	memset (&rc.server_addr, 0, sizeof (rc.server_addr));

	rc.server_addr.sin_port        = htons (atoi (options.local_port));
	rc.server_addr.sin_family      = AF_INET;
	rc.server_addr.sin_addr.s_addr = INADDR_ANY;

	rc.server_socket = socket (AF_INET, SOCK_STREAM, 0);
	if (rc.server_socket < OK)
	{
		perror ("build_server: socket ()");
		return KO;
	}

	if (bind (rc.server_socket, (struct sockaddr *) &rc.server_addr, sizeof (rc.server_addr)) < OK)
	{
		perror ("build_server: bind ()");
		return KO;
	}

	if (listen (rc.server_socket, 1) < OK)
	{
		perror ("build_server: listen ()");
		return KO;
	}

	return OK;
}

int wait_for_clients (void)
{
	int client_addr_size = sizeof (struct sockaddr_in);

	rc.client_socket = accept (rc.server_socket, (struct sockaddr *) &rc.client_addr, &client_addr_size);
	if (rc.client_socket < OK)
	{
		perror ("wait_for_clients: accept ()");
		return KO;
	}

	if (settings.log_to_stdout)
	{
		printf ("> %s tcptunnel: request from %s\n", get_current_timestamp (), inet_ntoa (rc.client_addr.sin_addr));
	}

	return OK;
}

int build_tunnel (void)
{
	rc.remote_host = gethostbyname (options.remote_host);
	if (rc.remote_host == NULL)
	{
		perror ("build_tunnel: gethostbyname ()");
		return KO;
	}

	memset (&rc.remote_addr, 0, sizeof (rc.remote_addr));

	rc.remote_addr.sin_family = AF_INET;
	rc.remote_addr.sin_port   = htons (atoi (options.remote_port));

	memcpy (&rc.remote_addr.sin_addr.s_addr, rc.remote_host->h_addr, rc.remote_host->h_length);

	rc.remote_socket = socket (AF_INET, SOCK_STREAM, 0);
	if (rc.remote_socket < OK)
	{
		perror ("build_tunnel: socket ()");
		return KO;
	}

	if (connect (rc.remote_socket, (struct sockaddr *) &rc.remote_addr, sizeof (rc.remote_addr)) < OK)
	{
		perror ("build_tunnel: connect ()");
		return KO;
	}

	return OK;
}

int use_tunnel (void)
{
	fd_set io;
	char buffer[SIZE];

	for (;;)
	{
		FD_ZERO (&io);
		FD_SET (rc.client_socket, &io);
		FD_SET (rc.remote_socket, &io);

		memset (buffer, 0, SIZE);

		if (select ( ( rc.client_socket < rc.remote_socket ?
				( rc.remote_socket < rc.server_socket ?
					rc.server_socket + 1 :
					rc.remote_socket + 1 ) :
				( rc.client_socket < rc.server_socket ?
					rc.server_socket + 1 :
					rc.client_socket + 1 ) ),
				&io, NULL, NULL, NULL) < 0 )
		{
			perror ("use_tunnel: select ()");
			break;
		}

		if (FD_ISSET (rc.client_socket, &io))
		{
			int count = recv (rc.client_socket, buffer, SIZE, 0);
			if (count < 0)
			{
				perror ("use_tunnel: client: recv ()");

				close (rc.client_socket);
				close (rc.remote_socket);

				return KO;
			}
			else if (count == 0)
			{
				perror ("use_tunnel: client: recv (): eof");

				close (rc.client_socket);
				close (rc.remote_socket);

				return KO;
			}
			send (rc.remote_socket, buffer, count, 0);

			if (settings.log_to_stdout)
			{
				printf ("> %s > ", get_current_timestamp ());
				fwrite (buffer, sizeof (char), count, stdout);
				fflush (stdout);
			}
		}

		if (FD_ISSET (rc.remote_socket, &io))
		{
			int count = recv (rc.remote_socket, buffer, SIZE, 0);
			if (count < 0)
			{
				perror ("use_tunnel: remote: recv ()");

				close (rc.client_socket);
				close (rc.remote_socket);

				return KO;
			}
			else if (count == 0)
			{
				perror ("use_tunnel: remote: recv (): eof");

				close (rc.client_socket);
				close (rc.remote_socket);

				return KO;
			}
			send (rc.client_socket, buffer, count, 0);

			if (settings.log_to_stdout)
			{
				fwrite (buffer, sizeof (char), count, stdout);
				fflush (stdout);
			}
		}
	}

	return OK;
}

void set_option (char **option, char *value)
{
	*option = (char *) malloc (sizeof (char) * (strlen (value) + 1));
	if (option == NULL)
	{
		perror ("set_option: malloc ()");
		exit (KO);
	}
	strcpy (*option, value);
}

char *get_current_timestamp (void)
{
	static char date_str[SIZE];
	time_t date;

	time (&date);
	strftime (date_str, SIZE - 1, "%a %d %H:%M:%S", localtime (&date));

	return date_str;
}

void print_usage (void)
{
	fprintf (stderr, "Usage: %s [options]\n\n", exec_name);
}

void print_helpinfo (void)
{
	fprintf (stderr, "Try `%s --help' for more options\n", exec_name);
}

void print_help (void)
{
	fprintf (stderr, "\
Options:\n\
  --version\n\
  --help               this help\n\n\
  --local-port=PORT    port to redirect\n\
  --remote-port=PORT   target port\n\
  --remote-host=HOST   target host\n\n\
  --stay-alive\n\
  --log-to-stdout\n\
\n");
}

void print_version (void)
{
	fprintf (stderr, "\
tcptunnel v" VERSION " Copyright (C) 2000-2004 Clemens Fuchslocher\n\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\n\
Written by Clemens Fuchslocher <clemens@allesdurcheinander.de>\n\
");
}

void print_missing (const char *message)
{
	print_usage ();
	fprintf (stderr, "%s: %s\n", exec_name, message);
	print_helpinfo ();
}
